/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.io;

import jalview.datamodel.AlignmentI;
import jalview.datamodel.SequenceI;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Map;

/**
 * Render associated attributes of an alignment. The heart of this code was
 * refactored from jalview.gui.AlignFrame and jalview.appletgui.AlignFrame TODO:
 * consider extending the html renderer to annotate elements with CSS ids
 * enabling finer output format control.
 * 
 */
public class AlignmentProperties
{
  private static final String BR_TAG = "<br>";

  private static final String NEWLINE = System.getProperty("line.separator");

  private static final String PCT_FORMAT = "%.1f%%";

  AlignmentI alignment;

  public AlignmentProperties(AlignmentI alignment)
  {
    this.alignment = alignment;
  }

  /**
   * render the alignment's properties report as text or an HTML fragment
   * 
   * @param html
   */
  protected StringBuilder writeProperties(boolean html)
  {
    StringBuilder sb = new StringBuilder(256);
    final String nl = html ? BR_TAG : NEWLINE;
    int totalLength = 0;
    int totalGaps = 0;
    int minLength = Integer.MAX_VALUE;
    int maxLength = 0;
    float maxUngapped = 0f;
    float minUngapped = Float.MAX_VALUE;

    final int height = alignment.getHeight();
    final int width = alignment.getWidth();

    for (int i = 0; i < height; i++)
    {
      SequenceI seq = alignment.getSequenceAt(i);
      // sequence length including gaps:
      int seqWidth = seq.getLength();
      // sequence length excluding gaps:
      int seqLength = 1 + seq.getEnd() - seq.getStart();
      int gapCount = seqWidth - seqLength; // includes padding
      totalLength += seqLength;
      totalGaps += gapCount;
      maxLength = Math.max(maxLength, seqLength);
      minLength = Math.min(minLength, seqLength);

      /*
       * proportion of aligned sequence that is ungapped
       * (note: normalised by alignment width here)
       */
      float ungapped = (seqWidth - gapCount) / (float) seqWidth;
      maxUngapped = Math.max(maxUngapped, ungapped);
      minUngapped = Math.min(minUngapped, ungapped);
    }
    float avgLength = totalLength / (float) height;

    sb.append(html ? "<table border=\"1\">" : nl);
    appendRow(sb, "Sequences", String.valueOf(height), html);
    appendRow(sb, "Alignment width", String.valueOf(width),
            html);
    appendRow(sb, "Minimum Sequence Length", String.valueOf(minLength),
            html);
    appendRow(sb, "Maximum Sequence Length", String.valueOf(maxLength),
            html);
    appendRow(sb, "Average Length", String.valueOf((int) avgLength), html);
    appendRow(sb, "Minimum (sequence length / width)",
            String.format(PCT_FORMAT, minUngapped * 100f), html);
    appendRow(sb, "Maximum (sequence length / width)",
            String.format(PCT_FORMAT, maxUngapped * 100f), html);
    appendRow(sb, "Residue density", String.format(PCT_FORMAT,
            (height * width - totalGaps) * 100f / (height * width)), html);
    appendRow(sb, "Gap density",
            String.format(PCT_FORMAT, totalGaps * 100f / (height * width)),
            html);
    sb.append(html ? "</table>" : nl);

    Map<Object, Object> props = alignment.getProperties();
    if (props != null && !props.isEmpty())
    {
      sb.append(nl);
      sb.append(nl);
      if (html)
      {
        sb.append("<table border=\"1\">");
      }

      /*
       * sort keys alphabetically for ease of reading the output
       */
      Object[] keys = props.keySet().toArray(new Object[props.size()]);
      Arrays.sort(keys, new Comparator<Object>()
      {
        @Override
        public int compare(Object o1, Object o2)
        {
          return String.CASE_INSENSITIVE_ORDER.compare(o1.toString(),
                  o2.toString());
        }
      });
      for (Object key : keys)
      {
        String value = props.get(key).toString();
        if (html)
        {
          value = value.replaceAll("\\R", value); // Java 8 newline matcher
          value = formatHrefs(value);
        }
        appendRow(sb, key.toString(), value, html);
      }
      if (html)
      {
        sb.append("</table>");
      }
    }
    return sb;
  }

  /**
   * Helper method to change any token starting with http into an html href
   * 
   * @param value
   * @return
   */
  private String formatHrefs(String value)
  {
    if (!value.contains("http"))
    {
      return value;
    }

    StringBuilder sb = new StringBuilder(value.length() * 3);
    String[] tokens = value.split("\\s");
    boolean found = false;
    boolean first = true;
    for (String token : tokens)
    {
      if (token.startsWith("http"))
      {
        token = "<a href=\"" + token + "\">" + token + "</a>";
        found = true;
      }
      if (!first)
      {
        sb.append(" ");
      }
      sb.append(token);
      first = false;
    }
    return found ? sb.toString() : value;
  }

  /**
   * A helper method to add one key-value row, optionally in HTML table entry
   * format
   * 
   * @param sb
   * @param key
   * @param value
   * @param html
   */
  private void appendRow(StringBuilder sb, String key, String value,
          boolean html)
  {
    if (html)
    {
      sb.append("<tr><td>").append(key).append("</td><td>").append(value)
              .append("</td></tr>");
    }
    else
    {
      sb.append(html ? BR_TAG : NEWLINE).append(key).append("\t")
              .append(value);
    }
  }

  /**
   * generate a report as plain text
   * 
   * @return
   */
  public StringBuilder formatAsString()
  {
    return writeProperties(false);
  }

  /**
   * generate a report as a fragment of html
   * 
   * @return
   */
  public StringBuilder formatAsHtml()
  {
    return writeProperties(true);
  }

}
